@@ -43,6 +43,7 @@
 #include "HTParse.h"
 #include "HTFormat.h"
 #include "HTTCP.h"
+#include "HTFile.h"		/* HTFileFormat() */
 
 /*		Hypertext object building machinery
 */
@@ -65,7 +66,7 @@ PRIVATE HTStructuredClass targetClass;		/* Its action routines */
 #define GOPHER_PROGRESS(foo) HTAlert(foo)
 
 
-#define NEXT_CHAR HTGetChararcter() 
+#define NEXT_CHAR HTInputSocket_getCharacter(isoc) 
 
 
 
@@ -146,7 +147,8 @@ PRIVATE void write_anchor ARGS2(CONST char *,text, CONST char *,addr)
 **
 */
 
-PRIVATE void parse_menu ARGS2 (
+PRIVATE void parse_menu ARGS3 (
+	int ,			s,
 	CONST char *,		arg,
 	HTParentAnchor *,	anAnchor)
 {
@@ -154,12 +156,14 @@ PRIVATE void parse_menu ARGS2 (
     char ch;
     char line[BIG];
     char address[BIG];
-    char *name, *selector;		/* Gopher menu fields */
-    char *host;
+    char *name = "";
+    char *selector = "";		/* Gopher menu fields */
+    char *host = "";
     char *port;
     char *p = line;
     CONST char *title;
-
+    HTInputSocket * isoc = HTInputSocket_new(s);
+    
 #define TAB 		'\t'
 #define HEX_ESCAPE 	'%'
 
@@ -233,7 +237,7 @@ PRIVATE void parse_menu ARGS2 (
 		    sprintf(address, "//%s/%c", host, gtype);
 		    q = address+ strlen(address);
 		    for(p=selector; *p; p++) {	/* Encode selector string */
-			if (acceptable[*p]) *q++ = *p;
+			if (acceptable[(int)*p]) *q++ = *p;
 			else {
 			    *q++ = HEX_ESCAPE;	/* Means hex coming */
 			    *q++ = hex[(TOASCII(*p)) >> 4];
@@ -266,8 +270,11 @@ PRIVATE void parse_menu ARGS2 (
     END(HTML_MENU);
     FREE_TARGET;
     
+    HTInputSocket_free(isoc);
     return;
 }
+
+
 /*	Parse a Gopher CSO document
  **	============================
  **
@@ -282,15 +289,17 @@ PRIVATE void parse_menu ARGS2 (
  **  secret@dxcern.cern.ch .
  */
 
-PRIVATE void parse_cso ARGS2 (
-			      CONST char *,	arg,
-			      HTParentAnchor *,anAnchor)
+PRIVATE void parse_cso ARGS3 (
+			int,	s,
+			CONST char *,		arg,
+			HTParentAnchor *,	anAnchor)
 {
     char ch;
     char line[BIG];
     char *p = line;
     char *second_colon, last_char='\0';
     CONST char *title;
+    HTInputSocket * isoc = HTInputSocket_new(s);
     
     title = HTAnchor_title(anAnchor);
     START(HTML_H1);
@@ -395,6 +404,7 @@ PRIVATE void parse_cso ARGS2 (
     END(HTML_PRE);
     PUTS("\n");
     FREE_TARGET;
+    HTInputSocket_free(isoc);
 
     return;  /* all done */
 } /* end of procedure */
@@ -484,17 +494,13 @@ PRIVATE void de_escape ARGS2(char *, command, CONST char *, selector)
 **	 Bug:	No decoding of strange data types as yet.
 **
 */
-PUBLIC int HTLoadGopher ARGS4(
-	CONST char *,		arg,
-	HTParentAnchor *,	anAnchor,
-	HTFormat,		format_out,
-	HTStream*,		sink)
+PUBLIC int HTLoadGopher ARGS1(HTRequest *, request)
 {
+    CONST char * arg = HTAnchor_physical(request->anchor);
     char *command;			/* The whole command */
     int status;				/* tcp return */
     char gtype;				/* Gopher Node type */
     char * selector;			/* Selector string */
- 
     struct sockaddr_in soc_address;	/* Binary network address */
     struct sockaddr_in* sin = &soc_address;
     
@@ -531,12 +537,14 @@ PUBLIC int HTLoadGopher ARGS4(
 	}
 	if (gtype == GOPHER_INDEX) {
 	    char * query;
-            HTAnchor_setIndex(anAnchor);	/* Search is allowed */
+            HTAnchor_setIndex(request->anchor);	/* Search is allowed */
 	    query = strchr(selector, '?');	/* Look for search string */
 	    if (!query || !query[1]) {		/* No search required */
-		target = HTML_new(anAnchor, format_out, sink);
+		target = HTML_new(request, NULL, WWW_HTML,
+			request->output_format, request->output_stream);
 		targetClass = *target->isa;
-		display_index(arg, anAnchor);	/* Display "cover page" */
+		display_index(arg, request->anchor);	/* Display "cover page" */
+		free(p1);		/* Leak fixed Henrik 27 Feb 94 */
 		return HT_LOADED;		/* Local function only */
 	    }
 	    *query++ = 0;			/* Skip '?' 	*/
@@ -556,12 +564,14 @@ PUBLIC int HTLoadGopher ARGS4(
 	    strcat(command, query);
         } else if (gtype == GOPHER_CSO) {
             char * query;
-            HTAnchor_setIndex(anAnchor);        /* Search is allowed */
+            HTAnchor_setIndex(request->anchor);        /* Search is allowed */
             query = strchr(selector, '?');      /* Look for search string */
             if (!query || !query[1]) {          /* No search required */
-		target = HTML_new(anAnchor, format_out, sink);
+		target = HTML_new(request, NULL, WWW_HTML,
+			request->output_format, request->output_stream);
 		targetClass = *target->isa;
-                display_cso(arg, anAnchor);     /* Display "cover page" */
+                display_cso(arg, request->anchor);     /* Display "cover page" */
+		free(p1);		/* Leak fixed Henrik 27 Feb 94 */
                 return HT_LOADED;                 /* Local function only */
             }
             *query++ = 0;                       /* Skip '?'     */
@@ -607,7 +617,6 @@ PUBLIC int HTLoadGopher ARGS4(
 	return HTInetStatus("connect");
     }
     
-    HTInitInput(s);		/* Set up input buffering */
     
     if (TRACE) fprintf(stderr, "HTGopher: Connected, writing command `%s' to socket %d\n", command, s);
     
@@ -632,42 +641,57 @@ PUBLIC int HTLoadGopher ARGS4(
     switch (gtype) {
     
     case GOPHER_HTML :
-    	HTParseSocket(WWW_HTML, format_out, anAnchor, s, sink);
+    	HTParseSocket(WWW_HTML,  s, request);
 	break;
 
     case GOPHER_GIF:
     case GOPHER_IMAGE:
-    	HTParseSocket(HTAtom_for("image/gif"), 
-			   format_out, anAnchor, s, sink);
+    	HTParseSocket(HTAtom_for("image/gif"), s, request);
   	break;
     case GOPHER_MENU :
     case GOPHER_INDEX :
-	target = HTML_new(anAnchor, format_out, sink);
+	target = HTML_new(request, NULL, WWW_HTML,
+			request->output_format, request->output_stream);
 	targetClass = *target->isa;
-        parse_menu(arg, anAnchor);
+        parse_menu(s,arg, request->anchor);
 	break;
 	 
     case GOPHER_CSO:
-	target = HTML_new(anAnchor, format_out, sink);
+	target = HTML_new(request, NULL, WWW_HTML,
+			request->output_format, request->output_stream);
 	targetClass = *target->isa;
-      	parse_cso(arg, anAnchor);
+      	parse_cso(s, arg, request->anchor);
 	break;
    	
       case GOPHER_MACBINHEX:
       case GOPHER_PCBINHEX:
       case GOPHER_UUENCODED:
       case GOPHER_BINARY:
-        /* Specifying WWW_UNKNOWN forces dump to local disk. */
-        HTParseSocket (WWW_UNKNOWN, format_out, anAnchor, s, sink);
+	{	/* Do our own filetyping -- maybe we get lucky */
+	    HTFormat format = HTFileFormat(arg,
+					   &request->content_encoding,
+					   &request->content_language);
+	    if (format) {
+		CTRACE(stderr,
+		       "Gopher...... figured out content-type myself: %s\n",
+		       HTAtom_name(format));
+		HTParseSocket(format, s, request);
+	    }
+	    else {
+		CTRACE(stderr,"Gopher...... using www/unknown\n");
+		/* Specifying WWW_UNKNOWN forces dump to local disk. */
+		HTParseSocket (WWW_UNKNOWN, s, request);
+	    }
+	}
 	break;
 
     case GOPHER_TEXT :
     default:			/* @@ parse as plain text */
-     	HTParseSocket(WWW_PLAINTEXT, format_out, anAnchor, s, sink);
+     	HTParseSocket(WWW_PLAINTEXT, s, request);
 	break;
 
       case GOPHER_SOUND :
-    	HTParseSocket(WWW_AUDIO, format_out, anAnchor, s, sink);
+    	HTParseSocket(WWW_AUDIO,  s, request);
 	break;
 	
     } /* switch(gtype) */
@@ -676,5 +700,5 @@ PUBLIC int HTLoadGopher ARGS4(
     return HT_LOADED;
 }
 
-GLOBALDEF PUBLIC HTProtocol HTGopher = { "gopher", HTLoadGopher, NULL };
+GLOBALDEF PUBLIC HTProtocol HTGopher = { "gopher", HTLoadGopher, NULL, NULL };
 
