@@ -9,6 +9,9 @@
 **	   Apr 91	vms-vms access included using DECnet syntax
 **	26 Jun 92 (JFG) When running over DECnet, suppressed FTP.
 **			Fixed access bug for relative names on VMS.
+**	   Sep 93 (MD)  Access to VMS files allows sharing.
+**	15 Nov 93 (MD)	Moved HTVMSname to HTVMSUTILS.C
+**	22 Feb 94 (MD)  Excluded two routines if we are not READING directories
 **
 ** Bugs:
 **	FTP: Cannot access VMS files from a unix machine.
@@ -16,18 +19,24 @@
 **	target machine runs VMS?
 */
 
-#include "HTMLDTD.h"            /* include the local DTD first.  If this
-                                   line isn't here. The compiler will try
-                                   and grabed the DTD that comes with the
-                                   library, which is different */
 #include "HTFile.h"		/* Implemented here */
-
+#include "TkWWWCmds.h"
 
 #define INFINITY 512		/* file name length @@ FIXME */
-#define MULTI_SUFFIX ".multi"   /* Extension for scanning formats */
+
+#ifdef VMS
+PRIVATE char * suffix_separators = "._";
+#else
+PRIVATE char * suffix_separators = ".,_";
+#endif
+
 
 #include "HTUtils.h"
 
+#ifdef VMS
+#include "HTVMSUtils.h"
+#endif /* VMS */
+
 #include "HTParse.h"
 #include "tcp.h"
 #include "HTTCP.h"
@@ -40,17 +49,19 @@
 #include "HTFWriter.h"
 #include "HTInit.h"
 #include "HTBTree.h"
-
-#include "TkWWWCmds.h"
+#include "HTFormat.h"
+#include "HTMulti.h"
 
 typedef struct _HTSuffix {
 	char *		suffix;
-	HTAtom *	rep;
-	HTAtom *	encoding;
+	HTAtom *	rep;		/* Content-Type */
+	HTAtom *	encoding;	/* Content-Encoding */
+	HTAtom *	language;	/* Content-Language */
 	float		quality;
 } HTSuffix;
 
 
+
 #ifdef USE_DIRENT		/* Set this for Sys V systems */
 #define STRUCT_DIRENT struct dirent
 #else
@@ -76,10 +87,12 @@ struct _HTStructured {
 
 PUBLIC int HTDirAccess = HT_DIR_OK;
 PUBLIC int HTDirReadme = HT_DIR_README_TOP;
+PUBLIC BOOL HTTakeBackup = YES;
+PUBLIC BOOL HTSuffixCaseSense = NO;	/* Are suffixes case sensitive */
 
 PRIVATE char *HTMountRoot = "/Net/";		/* Where to find mounts */
-#ifdef vms
-PRIVATE char *HTCacheRoot = "/WWW$SCRATCH/";   /* Where to cache things */
+#ifdef VMS
+PRIVATE char *HTCacheRoot = "/WWW$SCRATCH";   /* Where to cache things */
 #else
 PRIVATE char *HTCacheRoot = "/tmp/W3_Cache_";   /* Where to cache things */
 #endif
@@ -91,8 +104,8 @@ PRIVATE char *HTCacheRoot = "/tmp/W3_Cache_";   /* Where to cache things */
 */
 
 PRIVATE HTList * HTSuffixes = 0;
-PRIVATE HTSuffix no_suffix = { "*", NULL, NULL, 1.0 };
-PRIVATE HTSuffix unknown_suffix = { "*.*", NULL, NULL, 1.0};
+PRIVATE HTSuffix no_suffix = { "*", NULL, NULL, NULL, 1.0 };
+PRIVATE HTSuffix unknown_suffix = { "*.*", NULL, NULL, NULL, 1.0};
 
 
 /*	Define the representation associated with a file suffix
@@ -102,110 +115,169 @@ PRIVATE HTSuffix unknown_suffix = { "*.*", NULL, NULL, 1.0};
 **	representation.
 **	Calling this with suffix set to "*.*" will set the default
 **	representation for unknown suffix files which contain a ".".
+**
+**	If filename suffix is already defined its previous
+**	definition is overridden.
 */
-PUBLIC void HTSetSuffix ARGS4(
-	CONST char *,	suffix,
-	CONST char *,	representation,
-	CONST char *,	encoding,
-	float,		value)
+PUBLIC void HTAddType ARGS4(CONST char *,	suffix,
+			    CONST char *,	representation,
+			    CONST char *,	encoding,
+			    float,		value)
+{
+    HTSetSuffix(suffix, representation, encoding, NULL, value);
+}
+
+
+PUBLIC void HTAddEncoding ARGS3(CONST char *,	suffix,
+				CONST char *,	encoding,
+				float,		value)
+{
+    HTSetSuffix(suffix, NULL, encoding, NULL, value);
+}
+
+
+PUBLIC void HTAddLanguage ARGS3(CONST char *,	suffix,
+				CONST char *,	language,
+				float,		value)
+{
+    HTSetSuffix(suffix, NULL, NULL, language, value);
+}
+
+
+PUBLIC void HTSetSuffix ARGS5(CONST char *,	suffix,
+			      CONST char *,	representation,
+			      CONST char *,	encoding,
+			      CONST char *,	language,
+			      float,		value)
 {
-    
     HTSuffix * suff;
-    
+
     if (strcmp(suffix, "*")==0) suff = &no_suffix;
     else if (strcmp(suffix, "*.*")==0) suff = &unknown_suffix;
     else {
-	suff = (HTSuffix*) calloc(1, sizeof(HTSuffix));
-	if (suff == NULL) outofmem(__FILE__, "HTSetSuffix");
-	
-	if (!HTSuffixes) HTSuffixes = HTList_new();
-	HTList_addObject(HTSuffixes, suff);
-	
-	StrAllocCopy(suff->suffix, suffix);
+	HTList *cur = HTSuffixes;
+
+	while (NULL != (suff = (HTSuffix*)HTList_nextObject(cur))) {
+	    if (suff->suffix && 0==strcmp(suff->suffix, suffix))
+		break;
+	}
+	if (!suff) { /* Not found -- create a new node */
+	    suff = (HTSuffix*) calloc(1, sizeof(HTSuffix));
+	    if (suff == NULL) outofmem(__FILE__, "HTSetSuffix");
+
+	    if (!HTSuffixes) HTSuffixes = HTList_new();
+	    HTList_addObject(HTSuffixes, suff);
+
+	    StrAllocCopy(suff->suffix, suffix);
+	}
     }
 
-    suff->rep = HTAtom_for(representation);
-    
-    {
-    	char * enc = NULL;
+    if (representation)
+	suff->rep = HTAtom_for(representation);
+    if (language)
+	suff->language = HTAtom_for(language);
+    if (encoding) {
+	char * enc = NULL;
 	char * p;
 	StrAllocCopy(enc, encoding);
 	for (p=enc; *p; p++) *p = TOLOWER(*p);
-	suff->encoding = HTAtom_for(encoding);
+	suff->encoding = HTAtom_for(enc);
+	free(enc);	/* Leak fixed AL 6 Feb 1994 */
     }
-    
+
     suff->quality = value;
 }
 
 
-
-
-#ifdef vms
-/*	Convert unix-style name into VMS name
-**	-------------------------------------
-**
-** Bug:	Returns pointer to static -- non-reentrant
-*/
-PRIVATE char * vms_name(CONST char * nn, CONST char * fn)
+PRIVATE BOOL is_separator ARGS1(char, ch)
 {
-
-/*	We try converting the filename into Files-11 syntax. That is, we assume
-**	first that the file is, like us, on a VMS node. We try remote
-**	(or local) DECnet access. Files-11, VMS, VAX and DECnet
-**	are trademarks of Digital Equipment Corporation. 
-**	The node is assumed to be local if the hostname WITHOUT DOMAIN
-**	matches the local one. @@@
-*/
-    static char vmsname[INFINITY];	/* returned */
-    char * filename = (char*)malloc(strlen(fn)+1);
-    char * nodename = (char*)malloc(strlen(nn)+2+1);	/* Copies to hack */
-    char *second;		/* 2nd slash */
-    char *last;			/* last slash */
-    
-    char * hostname = HTHostName();
-
-    if (!filename || !nodename) outofmem(__FILE__, "vms_name");
-    strcpy(filename, fn);
-    strcpy(nodename, "");	/* On same node? Yes if node names match */
-    {
-        char *p, *q;
-        for (p=hostname, q=nn; *p && *p!='.' && *q && *q!='.'; p++, q++){
-	    if (TOUPPER(*p)!=TOUPPER(*q)) {
-	        strcpy(nodename, nn);
-		q = strchr(nodename, '.');	/* Mismatch */
-		if (q) *q=0;			/* Chop domain */
-		strcat(nodename, "::");		/* Try decnet anyway */
-		break;
-	    }
-	}
-    }
-
-    second = strchr(filename+1, '/');		/* 2nd slash */
-    last = strrchr(filename, '/');	/* last slash */
-        
-    if (!second) {				/* Only one slash */
-	sprintf(vmsname, "%s%s", nodename, filename + 1);
-    } else if(second==last) {		/* Exactly two slashes */
-	*second = 0;		/* Split filename from disk */
-	sprintf(vmsname, "%s%s:%s", nodename, filename+1, second+1);
-	*second = '/';	/* restore */
-    } else { 				/* More than two slashes */
-	char * p;
-	*second = 0;		/* Split disk from directories */
-	*last = 0;		/* Split dir from filename */
-	sprintf(vmsname, "%s%s:[%s]%s",
-		nodename, filename+1, second+1, last+1);
-	*second = *last = '/';	/* restore filename */
-	for (p=strchr(vmsname, '['); *p!=']'; p++)
-	    if (*p=='/') *p='.';	/* Convert dir sep.  to dots */
-    }
-    free(nodename);
-    free(filename);
-    return vmsname;
+    if (strchr(suffix_separators, ch)) return YES;
+    else return NO;
 }
 
 
-#endif /* vms */
+/* PUBLIC						HTSplitFilename()
+**
+**	Split the filename to an array of suffixes.
+**	Return the number of parts placed to the array.
+**	Array should have MAX_SUFF+1 items.
+*/
+PUBLIC int HTSplitFilename ARGS2(char *,	s_str,
+				 char **,	s_arr)
+{
+    char * start = s_str;
+    char * end;
+    char save;
+    int i;
+
+    if (!s_str || !s_arr) return 0;
+
+    for (i=0; i < MAX_SUFF && *start; i++) {
+	for(end=start+1; *end && !is_separator(*end); end++);
+	save = *end;
+	*end = 0;
+	StrAllocCopy(s_arr[i], start);	/* Frees the previous value */
+	*end = save;
+	start = end;
+    }
+    FREE(s_arr[i]);	/* Terminating NULL */
+    return i;
+}
+
+
+
+PUBLIC HTContentDescription * HTGetContentDescription ARGS2(char **,	actual,
+							    int,	n)
+{
+    HTContentDescription * cd;
+    int i;
+
+#ifndef NO_INIT    
+    if (!HTSuffixes) HTFileInit();
+#endif
+
+    if (n < 2) return NULL;	/* No suffix */
+
+    cd = (HTContentDescription*)calloc(1, sizeof(HTContentDescription));
+    if (!cd) outofmem(__FILE__, "HTContentDescription");
+
+    cd->quality = 1.0;
+
+    for (i=n-1; i>0; i--) {
+	HTList * cur = HTSuffixes;
+	HTSuffix * suff;
+	BOOL found = NO;
+
+	CTRACE(stderr, "Searching... for suffix %d: \"%s\"\n", i, actual[i]);
+
+	while ((suff = (HTSuffix*)HTList_nextObject(cur))) {
+	    if ((HTSuffixCaseSense && !strcmp(suff->suffix, actual[i])) ||
+		(!HTSuffixCaseSense && !strcasecomp(suff->suffix, actual[i]))){
+
+		if (!cd->content_type)
+		    cd->content_type = suff->rep;
+		if (!cd->content_encoding)
+		    cd->content_encoding = suff->encoding;
+		if (!cd->content_language)
+		    cd->content_language = suff->language;
+		if (suff->quality > 0.0000001)
+		    cd->quality *= suff->quality;
+
+		found = YES;
+		break;
+	    }
+	}
+	if (!found) {
+	    if (i < n-1)
+		break;
+	    else {
+		free(cd);
+		return NULL;
+	    }
+	}
+    }
+    return cd;
+}
 
 
 
@@ -254,6 +326,7 @@ Bug removed thanks to joe@athena.mit.edu */
 	END(HTML_PRE);
 	fclose(fp);
     } 
+    free(readme_file_name);	/* Leak fixed AL 6 Feb 1994 */
 }
 #endif
 
@@ -314,8 +387,15 @@ PUBLIC char * HTLocalName ARGS1(CONST char *,name)
     char * path = HTParse(name, "", PARSE_PATH+PARSE_PUNCTUATION);
     
     HTUnEscape(path);	/* Interpret % signs */
-    
-    if (0==strcmp(access, "file")) {
+
+    if (HTImServer && !*access) {
+	free(access);
+	free(host);
+	CTRACE(stderr, "Local filename is \"%s\"\n", path);
+	return(path);
+    }
+
+    if (0==strcmp(access, "file")) { /* local file */
         free(access);	
 	if ((0==strcasecomp(host, HTHostName())) ||
 	    (0==strcasecomp(host, "localhost")) || !*host) {
@@ -335,7 +415,14 @@ PUBLIC char * HTLocalName ARGS1(CONST char *,name)
     } else {  /* other access */
 	char * result;
         CONST char * home =  (CONST char*)getenv("HOME");
+#ifdef VMS
+	if (!home) 
+	    home = HTCacheRoot; 
+	else
+   	    home = HTVMS_wwwName(home);
+#else /* not VMS */
 	if (!home) home = "/tmp"; 
+#endif /* not VMS */
 	result = (char *)malloc(
 		strlen(home)+strlen(access)+strlen(host)+strlen(path)+6+1);
       if (result == NULL) outofmem(__FILE__, "HTLocalName");
@@ -393,15 +480,13 @@ PUBLIC char * WWW_nameOfFile ARGS1 (CONST char *,name)
 PUBLIC CONST char * HTFileSuffix ARGS1(HTAtom*, rep)
 {
     HTSuffix * suff;
-    int n;
-    int i;
+    HTList * cur;
 
 #ifndef NO_INIT    
     if (!HTSuffixes) HTFileInit();
 #endif
-    n = HTList_count(HTSuffixes);
-    for(i=0; i<n; i++) {
-	suff = HTList_objectAt(HTSuffixes, i);
+    cur = HTSuffixes;
+    while ((suff = (HTSuffix*)HTList_nextObject(cur))) {
 	if (suff->rep == rep) {
 	    return suff->suffix;		/* OK -- found */
 	}
@@ -419,40 +504,69 @@ PUBLIC CONST char * HTFileSuffix ARGS1(HTAtom*, rep)
 **	It will handle for example  x.txt, x.txt,Z, x.Z
 */
 
-PUBLIC HTFormat HTFileFormat ARGS2 (
-			CONST char *,	filename,
-			HTAtom **,	pencoding)
-
+PUBLIC HTFormat HTFileFormat ARGS3(CONST char *,	filename,
+				   HTAtom **,		pencoding,
+				   HTAtom **,		planguage)
 {
-    HTSuffix * suff;
+    char * actual[MAX_SUFF+1];
     int n;
-    int i;
-    int lf = strlen(filename);
+    HTContentDescription * cd;
+    HTFormat content_type = NULL;
+
+    if (!filename) return WWW_BINARY;
 
 #ifndef NO_INIT    
     if (!HTSuffixes) HTFileInit();
 #endif
-    *pencoding = NULL;
-    n = HTList_count(HTSuffixes);
-    for(i=0; i<n; i++) {
-        int ls;
-	suff = HTList_objectAt(HTSuffixes, i);
-	ls = strlen(suff->suffix);
+
+    for (n=0; n<MAX_SUFF+1; n++)  actual[n] = NULL;
+    n = HTSplitFilename((char*)filename, actual);
+    cd = HTGetContentDescription(actual, n);
+    while(n-- > 0) if (actual[n]) free(actual[n]);
+
+    if (cd) {
+	if (pencoding) *pencoding = cd->content_encoding;
+	if (planguage) *planguage = cd->content_language;
+	content_type = cd->content_type;
+	free(cd);
+    }
+    else {
+	HTSuffix * suff = strchr(filename,'.') ?
+	    (unknown_suffix.rep ? &unknown_suffix : &no_suffix) :
+		&no_suffix;
+
+	if (pencoding) *pencoding = suff->encoding;
+	if (planguage) *planguage = suff->language;
+	content_type = suff->rep;
+    }
+
+    if (pencoding && !*pencoding)
+	*pencoding = HTAtom_for("binary");
+    return content_type ? content_type : WWW_BINARY;
+}
+
+#ifdef OLD_CODE
+    HTSuffix * suff;
+    HTList * cur;
+    int lf = strlen(filename);
+
+    if (pencoding) *pencoding = NULL;
+    cur = HTSuffixes;
+    while ((suff = (HTSuffix*)HTList_nextObject(cur))) {
+        int ls = strlen(suff->suffix);
 	if ((ls <= lf) && 0==strcmp(suff->suffix, filename + lf - ls)) {
-	    int j;
-	    *pencoding = suff->encoding;
+	    if (pencoding) *pencoding = suff->encoding;
 	    if (suff->rep) return suff->rep;		/* OK -- found */
-	    
-	    for(j=0; j<n; j++) {  /* Got encoding, need representation */
-		int ls2;
-		suff = HTList_objectAt(HTSuffixes, j);
-		ls2 = strlen(suff->suffix);
-		if ((ls <= lf) && 0==strncmp(
-			suff->suffix, filename + lf - ls -ls2, ls2)) {
+
+	    /* Got encoding, need representation */
+	    cur = HTSuffixes;
+	    while ((suff = (HTSuffix*)HTList_nextObject(cur))) {
+		int ls2 = strlen(suff->suffix);
+		if ((ls <= lf) &&
+		    0==strncmp(suff->suffix, filename + lf - ls -ls2, ls2)) {
 		    if (suff->rep) return suff->rep;
 		}
 	    }
-	    
 	}
     }
     
@@ -463,10 +577,12 @@ PUBLIC HTFormat HTFileFormat ARGS2 (
 	 : &no_suffix;
 	 
     /* set default encoding unless found with suffix already */
-    if (!*pencoding) *pencoding = suff->encoding ? suff->encoding
+    if (pencoding && !*pencoding) *pencoding = suff->encoding ? suff->encoding
 				    : HTAtom_for("binary");
     return suff->rep ? suff->rep : WWW_BINARY;
 }
+#endif /* OLD_CODE */
+
 
 
 /*	Determine value from file name
@@ -478,18 +594,15 @@ PUBLIC float HTFileValue ARGS1 (CONST char *,filename)
 
 {
     HTSuffix * suff;
-    int n;
-    int i;
+    HTList * cur;
     int lf = strlen(filename);
 
 #ifndef NO_INIT    
     if (!HTSuffixes) HTFileInit();
 #endif
-    n = HTList_count(HTSuffixes);
-    for(i=0; i<n; i++) {
-        int ls;
-	suff = HTList_objectAt(HTSuffixes, i);
-	ls = strlen(suff->suffix);
+    cur = HTSuffixes;
+    while ((suff = (HTSuffix*)HTList_nextObject(cur))) {
+        int ls = strlen(suff->suffix);
 	if ((ls <= lf) && 0==strcmp(suff->suffix, filename + lf - ls)) {
 	    if (TRACE) fprintf(stderr, "File: Value of %s is %.3f\n",
 			       filename, suff->quality);
@@ -511,7 +624,7 @@ PUBLIC float HTFileValue ARGS1 (CONST char *,filename)
 **	2.	Isn't there a quicker way?
 */
 
-#ifdef vms
+#ifdef VMS
 #define NO_GROUPS
 #endif
 #ifdef NO_UNIX_IO
@@ -526,7 +639,11 @@ PUBLIC BOOL HTEditable ARGS1 (CONST char *,filename)
 #ifdef NO_GROUPS
     return NO;		/* Safe answer till we find the correct algorithm */
 #else
-    int 	groups[NGROUPS];	
+#ifdef NeXT
+    int		groups[NGROUPS];
+#else
+    gid_t 	groups[NGROUPS];
+#endif	
     uid_t	myUid;
     int		ngroups;			/* The number of groups  */
     struct stat	fileStatus;
@@ -575,17 +692,55 @@ PUBLIC BOOL HTEditable ARGS1 (CONST char *,filename)
 **	The stream must be used for writing back the file.
 **	@@@ no backup done
 */
-PUBLIC HTStream * HTFileSaveStream ARGS1(HTParentAnchor *, anchor)
+PUBLIC HTStream * HTFileSaveStream ARGS1(HTRequest *, request)
 {
 
-    CONST char * addr = HTAnchor_address((HTAnchor*)anchor);
-    char *  localname = HTLocalName(addr);
+    CONST char * addr = HTAnchor_address((HTAnchor*)request->anchor);
+    char *  filename = HTLocalName(addr);
+    FILE* fp;
+
+/*	@ Introduce CVS handling here one day
+*/
+/*	Take a backup before we do anything silly   931205
+*/        
+    if (HTTakeBackup) {
+    	char * backup_filename = 0;
+	char * p = backup_filename;
+	backup_filename = malloc(strlen(filename)+2);
+	if (!backup_filename) outofmem(__FILE__, "taking backup");
+	strcpy(backup_filename, filename);
+	for(p=backup_filename+strlen(backup_filename);; p--) {
+	    if ((*p=='/') || (p<backup_filename)) {
+	        p[1]=',';		/* Insert comma after slash */
+		break;
+	    }
+	    p[1] = p[0];	/* Move up everything to the right of it */
+	}
+	
+	if ((fp=fopen(filename, "r"))) {		/* File exists */
+	    fclose(fp);
+	    if (TRACE) fprintf(stderr, "File `%s' exists\n", filename);
+	    if (remove(backup_filename)) {
+		if (TRACE) fprintf(stderr, "Backup file `%s' removed\n",
+				   backup_filename);
+	    }
+	    if (rename(filename, backup_filename)) {	/* != 0 => Failure */
+		if (TRACE) fprintf(stderr, "Rename `%s' to `%s' FAILED!\n",
+				   filename, backup_filename);
+	    } else {					/* Success */
+		if (TRACE) fprintf(stderr, "Renamed `%s' to `%s'\n", filename,
+				   backup_filename);
+	    }
+	}
+    	free(backup_filename);
+    } /* if take backup */    
     
-    FILE* fp = fopen(localname, "w");
-    if (!fp) return NULL;
-    
-    return HTFWriter_new(fp);
+    {
+        fp = fopen(filename, "w");
+        if (!fp) return NULL;
     
+    	return HTFWriter_new(fp, NO);
+    }
 }
 
 /*      Output one directory entry
@@ -599,8 +754,7 @@ PUBLIC void HTDirEntry ARGS3(HTStructured *, target,
     char * escaped = HTEscape(entry, URL_XPALPHAS);
 
     /* If empty tail, gives absolute ref below */
-    relative = (char*) malloc(
-			      strlen(tail) + strlen(escaped)+2);
+    relative = (char*) malloc(strlen(tail) + strlen(escaped)+2);
     if (relative == NULL) outofmem(__FILE__, "DirRead");
     sprintf(relative, "%s/%s", tail, escaped);
     HTStartAnchor(target, NULL, relative);
@@ -687,54 +841,83 @@ PUBLIC void HTDirTitles ARGS2(HTStructured *, target,
 **			HTLOADED	OK 
 **
 */
-PUBLIC int HTLoadFile ARGS4 (
-	CONST char *,		addr,
-	HTParentAnchor *,	anchor,
-	HTFormat,		format_out,
-	HTStream *,		sink
-)
+PUBLIC int HTLoadFile ARGS1 (HTRequest *, request)
+
 {
+    CONST char * addr;
     char * filename;
+    char * access;
     HTFormat format;
-    char * nodename = 0;
+    static char * nodename = 0;
     char * newname=0;	/* Simplified name of file */
-    HTAtom * encoding;	/* @@ not used yet */
+    HTAtom * encoding;
+    HTAtom * language;
+
+    if (!request  ||  !request->anchor)
+	return HT_INTERNAL;
+
+    FREE(nodename);	/* From prev call - Leak fixed AL 6 Feb 1994 */
     
 /*	Reduce the filename to a basic form (hopefully unique!)
 */
+    addr = HTAnchor_physical(request->anchor);
     StrAllocCopy(newname, addr);
     filename=HTParse(newname, "", PARSE_PATH|PARSE_PUNCTUATION);
     nodename=HTParse(newname, "", PARSE_HOST);
-    free(newname);
-    
-    format = HTFileFormat(filename, &encoding);
 
+    /* If access is ftp => go directly to ftp code (henrik 27/02-94) */
+    access = HTParse(newname, "", PARSE_ACCESS);
+    if(!strcmp("ftp", access)) {
+        FREE(newname);
+        FREE(access);
+	FREE(filename);		                         /* Not used anymore */
+        goto try_ftp;
+    } else {
+        FREE(newname);
+        FREE(access);
+    }
 
-#ifdef vms
+    format = HTFileFormat(filename, &encoding, &language);
+
+#ifdef VMS
 /* Assume that the file is in Unix-style syntax if it contains a '/'
    after the leading one @@ */
     {
+        FILE * fp;
 	char * vmsname = strchr(filename + 1, '/') ?
-	  vms_name(nodename, filename) : filename + 1;
-	fd = open(vmsname, O_RDONLY, 0);
+	  HTVMS_name(nodename, filename) : filename + 1;
+	fp = fopen(vmsname, "r", "shr=put", "shr=upd");
 	
 /*	If the file wasn't VMS syntax, then perhaps it is ultrix
 */
-	if (fd<0) {
+	if (!fp) {
 	    char ultrixname[INFINITY];
 	    if (TRACE) fprintf(stderr, "HTFile: Can't open as %s\n", vmsname);
 	    sprintf(ultrixname, "%s::\"%s\"", nodename, filename);
-	    fd = open(ultrixname, O_RDONLY, 0);
-	    if (fd<0) {
+  	    fp = fopen(ultrixname, "r", "shr=put", "shr=upd");
+	    if (!fp) {
 		if (TRACE) fprintf(stderr, 
 				   "HTFile: Can't open as %s\n", ultrixname);
 	    }
 	}
+        if (fp)
+        {
+	    if (HTEditable(vmsname)) {
+		HTAtom * put = HTAtom_for("PUT");
+		HTList * methods = HTAnchor_methods(request->anchor);
+		if (HTList_indexOf(methods, put) == (-1)) {
+	   	    HTList_addObject(methods, put);
+	        }
+	    }
+	    HTParseFile(format, fp, request);
+	    fclose(fp);
+            return HT_LOADED;
+        }  /* If successfull open */
     }
 #else
 
-    free(filename);
-    
+    FREE(filename);
+
 /*	For unix, we try to translate the name into the name of a transparently
 **	mounted file.
 **
@@ -746,7 +929,8 @@ PUBLIC int HTLoadFile ARGS4 (
     if (!HTSecure) {		/* try local file system */
 	char * localname = HTLocalName(addr);
 	struct stat dir_info;
-	
+	char * multi;
+
 #ifdef GOT_READ_DIR
 
 /*			  Multiformat handling
@@ -754,71 +938,22 @@ PUBLIC int HTLoadFile ARGS4 (
 **	If needed, scan directory to find a good file.
 **  Bug:  we don't stat the file to find the length
 */
-	if ( (strlen(localname) > strlen(MULTI_SUFFIX))
-	   && (0==strcmp(localname + strlen(localname) - strlen(MULTI_SUFFIX),
-	                  MULTI_SUFFIX))) {
-	    DIR *dp;
-
-	    STRUCT_DIRENT * dirbuf;
-	    float best = NO_VALUE_FOUND;	/* So far best is bad */
-	    HTFormat best_rep = NULL;	/* Set when rep found */
-	    STRUCT_DIRENT best_dirbuf;	/* Best dir entry so far */
-
-	    char * base = strrchr(localname, '/');
-	    int baselen;
-
-	    if (!base || base == localname) goto forget_multi;
-	    *base++ = 0;		/* Just got directory name */
-	    baselen = strlen(base)- strlen(MULTI_SUFFIX);
-	    base[baselen] = 0;	/* Chop off suffix */
-
-	    dp = opendir(localname);
-	    if (!dp) {
-forget_multi:
-		free(localname);
-		return HTLoadError(sink, 500,
-			"Multiformat: directory scan failed.");
-	    }
-	    
-	    while ((dirbuf = readdir(dp))!=0) {
-			/* while there are directory entries to be read */
-		if (dirbuf->d_ino == 0) continue;
-				/* if the entry is not being used, skip it */
-		
-		if (dirbuf->d_namlen > baselen &&      /* Match? */
-		    !strncmp(dirbuf->d_name, base, baselen)) {	
-		    HTFormat rep = HTFileFormat(dirbuf->d_name, &encoding);
-		    float value = HTStackValue(rep, format_out,
-		    				HTFileValue(dirbuf->d_name),
-						0.0  /* @@@@@@ */);
-		    if (value != NO_VALUE_FOUND) {
-		        if (TRACE) fprintf(stderr,
-				"HTFile: value of presenting %s is %f\n",
-				HTAtom_name(rep), value);
-			if  (value > best) {
-			    best_rep = rep;
-			    best = value;
-			    best_dirbuf = *dirbuf;
-		       }
-		    }	/* if best so far */ 		    
-		 } /* if match */  
-		    
-	    } /* end while directory entries left to read */
-	    closedir(dp);
-	    
-	    if (best_rep) {
-		format = best_rep;
-		base[-1] = '/';		/* Restore directory name */
-		base[0] = 0;
-		StrAllocCat(localname, best_dirbuf.d_name);
+	multi = strrchr(localname, MULTI_SUFFIX[0]);
+	if (multi && !strcmp(multi, MULTI_SUFFIX)) {
+	    struct stat stat_info;
+	    char * new_path = HTMulti(request, localname, &stat_info);
+	    if (new_path) {
+		FREE(localname);
+		localname = new_path;
+		HTAnchor_setPhysical(request->anchor, localname);
+		format = HTFileFormat(localname, &encoding, &language);
 		goto open_file;
-		
-	    } else { 			/* If not found suitable file */
-		free(localname);
-		return HTLoadError(sink, 403,	/* List formats? */
+	    }
+	    else { 			/* If not found suitable file */
+		FREE(localname);
+		return HTLoadError(request, 403,	/* List formats? */
 		   "Could not find suitable representation for transmission.");
 	    }
-	    /*NOTREACHED*/
 	} /* if multi suffix */
 /*
 **	Check to see if the 'localname' is in fact a directory.  If it is
@@ -837,7 +972,6 @@ forget_multi:
 
 	}  else {		/* Stat was OK */
 		
-
 	    if (((dir_info.st_mode) & S_IFMT) == S_IFDIR) {
 		/* if localname is a directory */	
 
@@ -848,13 +982,13 @@ forget_multi:
 		STRUCT_DIRENT * dirbuf;
 		
 		char * logical;
-		char * tail;
-		
+		static char * tail = NULL;
+
 		BOOL present[HTML_A_ATTRIBUTES];
 		
 		char * tmpfilename = NULL;
 		struct stat file_info;
-		
+
 		if (TRACE)
 		    fprintf(stderr,"%s is a directory\n",localname);
 			
@@ -863,8 +997,8 @@ forget_multi:
 **	marker file can be browsed
 */
 		if (HTDirAccess == HT_DIR_FORBID) {
-		    free(localname);
-		    return HTLoadError(sink, 403,
+		    FREE(localname);
+		    return HTLoadError(request, 403,
 		    "Directory browsing is not allowed.");
 		}
 
@@ -877,8 +1011,8 @@ forget_multi:
 		    strcat(enable_file_name, "/");
 		    strcat(enable_file_name, HT_DIR_ENABLE_FILE);
 		    if (stat(enable_file_name, &file_info) != 0) {
-			free(localname);
-			return HTLoadError(sink, 403,
+			FREE(localname);
+			return HTLoadError(request, 403,
 			"Selective access is not enabled for this directory");
 		    }
 		}
@@ -886,17 +1020,27 @@ forget_multi:
  
 		dp = opendir(localname);
 		if (!dp) {
-		    free(localname);
-		    return HTLoadError(sink, 403, "This directory is not readable.");
+		    FREE(localname);
+		    return HTLoadError(request, 403,
+				       "This directory is not readable.");
 		}
 
 
  /*	Directory access is allowed and possible
  */
-		logical = HTAnchor_address((HTAnchor*)anchor);
+		logical = HTAnchor_address((HTAnchor*)request->anchor);
+		FREE(tail);	/* From previous call */
 		tail = strrchr(logical, '/') +1;	/* last part or "" */
-		
-		target = HTML_new(anchor, format_out, sink);
+
+		/*
+		** Fix AL 26.1.94: make dir.indexing work also if
+		** there is a trailing slash:
+		*/
+		if (!tail || !*tail) tail = ".";
+		tail = HTEscape(tail, URL_XALPHAS);
+
+		target = HTML_new(request, NULL, WWW_HTML, 
+			request->output_format, request->output_stream);
 		targetClass = *target->isa;	/* Copy routine entry points */
 		    
   		{ int i;
@@ -904,7 +1048,7 @@ forget_multi:
 				present[i] = (i==HTML_A_HREF);
 		}
 		
-                HTDirTitles(target, (HTAnchor *)anchor);
+                HTDirTitles(target, (HTAnchor *)request->anchor);
 
                 if (HTDirReadme == HT_DIR_README_TOP)
 		    do_readme(target, localname);
@@ -1012,7 +1156,7 @@ forget_multi:
 		    if (HTDirReadme == HT_DIR_README_BOTTOM)
 			  do_readme(target, localname);
 		    FREE_TARGET;
-		    free(localname);
+		    FREE(localname);
 		    return HT_LOADED;	/* document loaded */
 		}
 
@@ -1026,35 +1170,53 @@ forget_multi:
 open_file:
 	{
 	    FILE * fp = fopen(localname,"r");
+
 	    if(TRACE) fprintf (stderr, "HTFile: Opening `%s' gives %p\n",
 				localname, (void*)fp);
 	    if (fp) {		/* Good! */
 		if (HTEditable(localname)) {
 		    HTAtom * put = HTAtom_for("PUT");
-		    HTList * methods = HTAnchor_methods(anchor);
+		    HTList * methods = HTAnchor_methods(request->anchor);
 		    if (HTList_indexOf(methods, put) == (-1)) {
 			HTList_addObject(methods, put);
 		    }
 		}
-		 if (HTTkUseInPlace(format, format_out)) 
-		   HTTkSetOutputFile(format, anchor, localname);
-		 else
-		   HTParseFile(format, format_out, anchor, fp, sink);
-		free(localname);
+		
+               if (HTTkUseInPlace(format, request)) 
+                  HTTkSetOutputFile(request, format,
+				    localname, encoding);
+                else
+                  HTParseFile(format, fp, request);
 		fclose(fp);
+		FREE(localname);
 		return HT_LOADED;
 	    }  /* If succesfull open */
+	    FREE(localname);		/* If error in fopen */
 	}    /* scope of fp */
     }  /* local unix file system */    
 #endif
 #endif
 
+  try_ftp:
+
 #ifndef DECNET
 /*	Now, as transparently mounted access has failed, we try FTP.
 */
     {
-	if (strcmp(nodename, HTHostName())!=0)
-	    return HTFTPLoad(addr, anchor, format_out, sink);
+#ifdef OLD_CODE
+	if (nodename && strcmp(nodename, HTHostName())!=0)
+	    return HTFTPLoad(request, NULL, addr,
+	    request->anchor, request->output_format, request->output_stream);
+#endif
+	char * newname = NULL;
+	StrAllocCopy(newname, "ftp:");
+	if (!strncmp(addr, "file:", 5))
+	    StrAllocCat(newname, addr+5);
+	else
+	    StrAllocCat(newname, addr);
+	HTAnchor_setPhysical(request->anchor, newname);
+	free(newname);
+	return HTLoad(request);
     }
 #endif
 
@@ -1062,8 +1224,8 @@ open_file:
 */
     {
     	if (TRACE)
-    	printf("Can't open `%s', errno=%d\n", addr, errno);
-	return HTLoadError(sink, 403, "Can't access requested file.");
+	    fprintf(stderr, "Can't open `%s', errno=%d\n", addr, errno);
+	return HTLoadError(request, 403, "Can't access requested file.");
     }
     
  
@@ -1071,5 +1233,6 @@ open_file:
 
 /*		Protocol descriptors
 */
-PUBLIC HTProtocol HTFTP  = { "ftp", HTLoadFile, 0 };
-PUBLIC HTProtocol HTFile = { "file", HTLoadFile, HTFileSaveStream };
+GLOBALDEF PUBLIC HTProtocol HTFTP  = { "ftp", HTFTPLoad, 0 , 0 };
+GLOBALDEF PUBLIC HTProtocol HTFile = { "file", HTLoadFile,
+					HTFileSaveStream, 0 };
