@@ -1,8 +1,10 @@
 #include <TkWWWCmds.h>
+#include <TkWWWBitmap.h>
 #include <tcl.h>
 #include <tk.h>
 
-int Tcl_AppInit(interp) 
+int 
+Tcl_AppInit(interp) 
      Tcl_Interp *interp;
 {
   Tk_Window main;
@@ -21,5 +23,9 @@ int Tcl_AppInit(interp)
     return TCL_ERROR;
   }
 
+  if (WWW_BitmapInit(interp) == TCL_ERROR) {
+    return TCL_ERROR;
+  }
+
   return (TCL_OK);
 }
