@@ -8,7 +8,7 @@ int Tcl_AppInit(interp)
     return TCL_ERROR;
   }
 
-  if (WWW_Init(interp) == TCL_ERROR) {
+  if (WWW_AppInit(interp) == TCL_ERROR) {
     return TCL_ERROR;
   }
 
